# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Classes that track game state.

Note that these classes track game state, but do not implement game rules.  The
only validations are to prevent changes that literally cannot be represented in
game state, such as selecting an invalid square.  All other rules (such as the
restriction that only one pawn can occupy a space, or the act of sliding down a
slider, etc.) are implemented in the rules module, using the methods available on these
classes.

In many cases, private attributes are accessible in the constructor to support
serialization and deserialization.  In general, callers should not pass in optional
constructor arguments and should not modify public attributes if an alternate method
is available for use.

Attributes:
    MIN_PLAYERS(int): Minimum number of players in a game
    MAX_PLAYERS(int): Maximum number of players in a game
    PAWNS(int): Number of pawns per player
    SAFE_SQUARES(int): Number of safe squares for each color
    BOARD_SQUARES(int): Number of squares around the outside of the board
    ADULT_HAND(int): Number of cards in a player's hand for an adult mode game
    DECK_COUNTS(Dict[Card, int]): Dictionary from card name to number of cards in a standard deck
    DECK_SIZE(int): The total expected size of a complete deck
    DRAW_AGAIN(Dict[Card, bool]): Whether each card results in a draw again action
    CIRCLE(Dict[PlayerColor, Position]): The position of the start circle for each color
    TURN(Dict[PlayerColor, Position()): The position of the turn square for each color, where forward movement turns into safe zone
    SLIDE(Dict[PlayerColor, Tuple(int, int)): The slide start/end squares for each color
"""

from __future__ import annotations  # see: https://stackoverflow.com/a/33533514/2907667

import json
import random
from enum import Enum
from typing import Dict, List, Optional

import pendulum
from attrs import define, field, frozen
from pendulum.datetime import DateTime

from .util import CattrConverter

# Cattr converter that serializes/deserializes DateTime to an ISO 8601 timestamp
_CONVERTER = CattrConverter()

# A game consists of 2-4 players
MIN_PLAYERS = 2
MAX_PLAYERS = 4

# There are 4 pawns per player, numbered 0-3
PAWNS = 4

# There are 5 safe squares for each color, numbered 0-4
SAFE_SQUARES = 5

# There are 60 squares around the outside of the board, numbered 0-59
BOARD_SQUARES = 60


class GameMode(Enum):
    """Available game play modes."""

    STANDARD = "Standard"
    ADULT = "Adult"


class PlayerColor(Enum):
    """Enumeration of all player colors, listed in order of use."""

    RED = "Red"
    YELLOW = "Yellow"
    GREEN = "Green"
    BLUE = "Blue"


class CardType(Enum):
    """
    All legal types of cards.

    The "A" card (CARD_APOLOGIES) is like the "Sorry" card in the original game.
    """

    CARD_1 = "1"
    CARD_2 = "2"
    CARD_3 = "3"
    CARD_4 = "4"
    CARD_5 = "5"
    CARD_7 = "7"
    CARD_8 = "8"
    CARD_10 = "10"
    CARD_11 = "11"
    CARD_12 = "12"
    CARD_APOLOGIES = "A"  # This is like the "Sorry" card in the original game


# For an adult-mode game, we deal out 5 cards
ADULT_HAND = 5

# Deck definitions
DECK_COUNTS = {
    CardType.CARD_1: 5,
    CardType.CARD_2: 4,
    CardType.CARD_3: 4,
    CardType.CARD_4: 4,
    CardType.CARD_5: 4,
    CardType.CARD_7: 4,
    CardType.CARD_8: 4,
    CardType.CARD_10: 4,
    CardType.CARD_11: 4,
    CardType.CARD_12: 4,
    CardType.CARD_APOLOGIES: 4,
}
DECK_SIZE = sum(DECK_COUNTS.values())

# Whether a card draws again
DRAW_AGAIN = {
    CardType.CARD_1: False,
    CardType.CARD_2: True,
    CardType.CARD_3: False,
    CardType.CARD_4: False,
    CardType.CARD_5: False,
    CardType.CARD_7: False,
    CardType.CARD_8: False,
    CardType.CARD_10: False,
    CardType.CARD_11: False,
    CardType.CARD_12: False,
    CardType.CARD_APOLOGIES: False,
}


@frozen(order=True)
class Card:
    # noinspection PyUnresolvedReferences
    """
    A card in a deck or in a player's hand.

    Attributes:
        id(str): Unique identifier for this card
        cardtype(CardType): The type of the card
    """

    id: str
    cardtype: CardType


@define(slots=False)
class Deck:
    """
    The deck of cards associated with a game.

    Callers should not pass in constructor arguments. These are accessible to
    support serialization and deserialization.
    """

    _draw_pile: Dict[str, Card] = field()
    _discard_pile: Dict[str, Card] = field(factory=dict)

    # noinspection PyUnresolvedReferences
    @_draw_pile.default
    def _default_draw_pile(self) -> Dict[str, Card]:
        pile = {}
        cardid = 0
        for card in CardType:
            for _ in range(DECK_COUNTS[card]):
                pile["%d" % cardid] = Card("%d" % cardid, card)
                cardid += 1
        return pile

    def draw(self) -> Card:
        """Draw a random card from the draw pile."""
        if len(self._draw_pile) < 1:
            # this is equivalent to shuffling the discard pile into the draw pile
            for card in list(self._discard_pile.values()):
                self._discard_pile.pop(card.id)
                self._draw_pile[card.id] = card
        if len(self._draw_pile) < 1:
            raise ValueError("No cards available in deck")  # in any normal game, this should never happen
        return self._draw_pile.pop(random.choice(list(self._draw_pile.keys())))

    def discard(self, card: Card) -> None:
        """Discard back to the discard pile."""
        if card.id in self._draw_pile or card.id in self._discard_pile:
            raise ValueError("Card already exists in deck")
        self._discard_pile[card.id] = card


@define(slots=False)
class Position:
    """
    The position of a pawn on the board.

    Callers should not pass in or directly modify the start, home, safe, or square
    attributes.  These are accessible to support serialization and deserialization.
    Instead, use the provided methods to safely modify the object in-place.

    Attributes:
        start(boolean): Whether this pawn resides in its start area
        home(boolean): Whether this pawn resides in its home area
        safe(int): Zero-based index of the square in the safe area where this pawn resides
        square(int): Zero-based index of the square on the board where this pawn resides
    """

    start: bool = True
    home: bool = False
    safe: Optional[int] = None
    square: Optional[int] = None

    def __str__(self) -> str:
        if self.home:
            return "home"
        elif self.start:
            return "start"
        elif self.safe is not None:
            return "safe %s" % self.safe
        else:
            return "square %s" % self.square

    def copy(self) -> Position:
        """Return a fully-independent copy of the position."""
        return _CONVERTER.structure(_CONVERTER.unstructure(self), Position)

    def move_to_position(self, position: Position) -> Position:
        """
        Move the pawn to a specific position on the board.

        Returns:
            Position: A reference to the position, for chaining

        Raises:
            ValueError: If the position is invalid
        """
        fields = 0
        if position.start:
            fields += 1
        if position.home:
            fields += 1
        if position.safe is not None:
            fields += 1
        if position.square is not None:
            fields += 1
        if fields != 1:
            raise ValueError("Invalid position")
        if position.start:
            self.move_to_start()
        elif position.home:
            self.move_to_home()
        elif position.safe is not None:
            self.move_to_safe(position.safe)
        elif position.square is not None:
            self.move_to_square(position.square)
        return self

    def move_to_start(self) -> Position:
        """
        Move the pawn back to its start area.

        Returns:
            Position: A reference to the position, for chaining

        Raises:
            ValueError: If the position is invalid
        """
        self.start = True
        self.home = False
        self.safe = None
        self.square = None
        return self

    def move_to_home(self) -> Position:
        """
        Move the pawn to its home area.

        Returns:
            Position: A reference to the position, for chaining

        Raises:
            ValueError: If the position is invalid
        """
        self.start = False
        self.home = True
        self.safe = None
        self.square = None
        return self

    def move_to_safe(self, square: int) -> Position:
        """
        Move the pawn to a square in its safe area.

        Args:
            square(int): Zero-based index of the square in the safe area

        Returns:
            Position: A reference to the position, for chaining

        Raises:
            ValueError: If the square is not valid
        """
        if square not in range(SAFE_SQUARES):
            raise ValueError("Invalid square")
        self.start = False
        self.home = False
        self.safe = square
        self.square = None
        return self

    def move_to_square(self, square: int) -> Position:
        """
        Move the pawn to a square on the board.

        Args:
            square(int): Zero-based index of the square on the board where this pawn resides

        Returns:
            Position: A reference to the position, for chaining

        Raises:
            ValueError: If the square is not valid
        """
        if square not in range(BOARD_SQUARES):
            raise ValueError("Invalid square")
        self.start = False
        self.home = False
        self.safe = None
        self.square = square
        return self


@define
class Pawn:
    # noinspection PyUnresolvedReferences
    """
    A pawn on the board, belonging to a player.

    Callers should not pass in the position attribute.  This is accessible
    to support serialization and deserialization. Instead, use the provided
    methods to safely modify the position in-place.

    Attributes:
        color(str): The color of this pawn
        index(int): Zero-based index of this pawn for a given user
        name(str): The full name of this pawn as "colorindex"
        position(Position): The position of this pawn on the board
    """

    color: PlayerColor
    index: int
    name: str = field()
    position: Position = field()

    # noinspection PyUnresolvedReferences
    @name.default
    def _default_name(self) -> str:
        return "%s%s" % (self.color.value, self.index)

    @position.default
    def _default_position(self) -> Position:
        return Position()

    def __str__(self) -> str:
        return "%s->%s" % (self.name, self.position)


@define
class Player:
    # noinspection PyUnresolvedReferences
    """
    A player, which has a color and a set of pawns.

    Callers should not pass in the hand and pawns constructor arguments.  These
    are accessible to support serialization and deserialization.

    Attributes:
        color(str): The color of the player
        hand(List[Card]): List of cards in the player's hand
        pawns(List[Pawn]): List of all pawns belonging to the player
        turns(int): Number of turns for this player
    """

    color: PlayerColor
    hand: List[Card] = field(factory=list)
    pawns: List[Pawn] = field()
    turns: int = 0

    # noinspection PyUnresolvedReferences
    @pawns.default
    def _default_pawns(self) -> List[Pawn]:
        return [Pawn(self.color, index) for index in range(0, PAWNS)]

    def copy(self) -> Player:
        """Return a fully-independent copy of the player."""
        return _CONVERTER.structure(_CONVERTER.unstructure(self), Player)

    def public_data(self) -> Player:
        """Return a fully-independent copy of the player with only public data visible."""
        player = self.copy()
        del player.hand[:]  # other players should not see this player's hand when making decisions
        return player

    def find_first_pawn_in_start(self) -> Optional[Pawn]:
        """Find the first pawn in the start area, if any."""
        for pawn in self.pawns:
            if pawn.position.start:
                return pawn
        return None

    def all_pawns_in_home(self) -> bool:
        """Whether all of this user's pawns are in home."""
        for pawn in self.pawns:
            if not pawn.position.home:
                return False
        return True


@define
class History:
    # noinspection PyUnresolvedReferences
    """
    Tracks an action taken during the game.

    Attributes:
        action(str): String describing the action
        color(Optional[PlayerColor]): Color of the player associated with the action
        card(Optional[CardType]): Card associated with the action
        timestamp(DateTime): Timestamp tied to the action (defaults to current time)
    """

    action: str
    color: Optional[PlayerColor] = None
    card: Optional[CardType] = None
    timestamp: DateTime = field()

    # noinspection PyUnresolvedReferences
    @timestamp.default
    def _default_timestamp(self) -> DateTime:
        return pendulum.now(pendulum.UTC)  # type: ignore[attr-defined]

    def __str__(self) -> str:
        time = self.timestamp.to_time_string()  # type: ignore
        color = "General" if not self.color else self.color.value
        action = self.action
        return "[%s] %s - %s" % (time, color, action)


@define
class PlayerView:
    # noinspection PyUnresolvedReferences
    """
    A player-specific view of the game, showing only the information a player would have available on their turn.

    Attributes:
        player(Player): The player associated with the view.
        opponents(Dict[PlayerColor, Player]): The player's opponents, with private information stripped
    """

    player: Player
    opponents: Dict[PlayerColor, Player]

    def copy(self) -> PlayerView:
        """Return a fully-independent copy of the player view."""
        return _CONVERTER.structure(_CONVERTER.unstructure(self), PlayerView)

    def get_pawn(self, prototype: Pawn) -> Optional[Pawn]:
        """Return the pawn from this view with the same color and index."""
        for pawn in self.all_pawns():
            if pawn.color == prototype.color and pawn.index == prototype.index:
                return pawn
        return None

    def all_pawns(self) -> List[Pawn]:
        """Return a list of all pawns on the board."""
        pawns = []
        pawns.extend(self.player.pawns)
        for opponent in self.opponents.values():
            pawns.extend(opponent.pawns)
        return pawns


@define(slots=False)
class Game:
    # noinspection PyUnresolvedReferences
    """
    The game, consisting of state for a set of players.

    Callers should not pass in optional constructor arguments.  These are accessible
    to support serialization and deserialization.

    Attributes:
        playercount(int): Number of players in the game
        players(Dict[PlayerColor, Player]): All players in the game
        deck(Deck): The deck of cards for the game
        history(History): Game history
    """

    playercount: int = field()
    players: Dict[PlayerColor, Player] = field()
    deck: Deck = field(factory=Deck)
    history: List[History] = field(factory=list)

    # noinspection PyUnresolvedReferences
    @playercount.validator
    def _check_playercount(self, _attribute: str, value: int) -> None:
        if value < MIN_PLAYERS or value > MAX_PLAYERS:
            raise ValueError("Invalid number of players")

    # noinspection PyUnresolvedReferences
    @players.default
    def _default_players(self) -> Dict[PlayerColor, Player]:
        return {color: Player(color) for color in list(PlayerColor)[: self.playercount]}

    @property
    def started(self) -> bool:
        """Whether the game has been started."""
        return len(self.history) > 0  # if there is any history the game has been started

    @property
    def completed(self) -> bool:
        """Whether the game is completed."""
        for player in self.players.values():
            if player.all_pawns_in_home():
                return True
        return False

    @property
    def winner(self) -> Optional[Player]:
        """The winner of the game, if any."""
        for player in self.players.values():
            if player.all_pawns_in_home():
                return player
        return None

    def copy(self) -> Game:
        """Return a fully-independent copy of the game."""
        return _CONVERTER.structure(_CONVERTER.unstructure(self), Game)

    def to_json(self) -> str:
        """Serialize the game state to JSON."""
        return json.dumps(_CONVERTER.unstructure(self), indent="  ")

    @staticmethod
    def from_json(data: str) -> Game:
        """Deserialize the game state from JSON."""
        return _CONVERTER.structure(json.loads(data), Game)

    def track(self, action: str, player: Optional[Player] = None, card: Optional[Card] = None) -> None:
        """Tracks an action taken during the game."""
        self.history.append(History(action, player.color if player else None, card.cardtype if card else None))
        if player:
            self.players[player.color].turns += 1

    def create_player_view(self, color: PlayerColor) -> PlayerView:
        """Return a player-specific view of the game, showing only the information a player would have available on their turn."""
        player = self.players[color].copy()
        opponents = {player.color: player.public_data() for player in self.players.values() if player.color != color}
        return PlayerView(player, opponents)


# The start circles for each color
CIRCLE = {
    PlayerColor.RED: Position().move_to_square(4),
    PlayerColor.BLUE: Position().move_to_square(19),
    PlayerColor.YELLOW: Position().move_to_square(34),
    PlayerColor.GREEN: Position().move_to_square(49),
}

# The turn squares for each color, where forward movement turns into the safe zone
TURN = {
    PlayerColor.RED: Position().move_to_square(2),
    PlayerColor.BLUE: Position().move_to_square(17),
    PlayerColor.YELLOW: Position().move_to_square(32),
    PlayerColor.GREEN: Position().move_to_square(47),
}

# The slide start/end positions for each color
SLIDE = {
    PlayerColor.RED: ((1, 4), (9, 13)),
    PlayerColor.BLUE: ((16, 19), (24, 28)),
    PlayerColor.YELLOW: ((31, 34), (39, 43)),
    PlayerColor.GREEN: ((46, 49), (54, 58)),
}



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

# We'd prefer to disable these only on Windows, but that isn't an option
# pylint: disable=bad-option-value,no-member,no-name-in-module

"""
Implements a quick'n'dirty game-playing demo using curses.
"""

import curses
import sys
from curses import endwin
from signal import SIGWINCH, signal
from time import sleep

from .engine import Character, Engine
from .game import GameMode
from .render import render_board
from .source import CharacterInputSource

# Minimum terminal size needed to support the demo
_MIN_COLS = 155
_MIN_ROWS = 58


class TerminalSizeError(Exception):
    def __init__(self, msg):
        self.msg = msg


def _render_hand(player):
    """Return a string describing the cards in a player's hand."""
    if not player.hand:
        return "n/a"
    else:
        return "%s" % [card.cardtype.value for card in sorted(player.hand)]


def _draw(stdscr, board, state, history):
    """Draw the static portions of the screen."""
    stdscr.clear()
    stdscr.border()
    stdscr.refresh()

    board.border()
    board.refresh()

    state.border()
    state.refresh()

    history.border()
    history.refresh()


def _refresh(source, engine, game, delay_sec, stdscr, board, state, history):
    """Refresh the dynamic portions of the screen."""
    _refresh_screen(source, engine, delay_sec, game, stdscr)
    _refresh_board(game, board)
    _refresh_state(source, engine, delay_sec, game, state)
    _refresh_history(game, history)


def _refresh_screen(unused_source, unused_engine, unused_game, unused_delay_sec, stdscr):
    stdscr.border()
    stdscr.addstr(1, 95, "APOLOGIES DEMO")
    stdscr.addstr(1, 138, "CTRL-C TO EXIT")
    stdscr.move(_MIN_ROWS - 2, _MIN_COLS - 2)  # bottom-right corner
    stdscr.refresh()


def _refresh_board(game, board):
    """Refresh the game board display section of the screen."""
    board.clear()
    board.border()

    row = 0
    for line in render_board(game).splitlines():
        board.addstr(row, 1, line)
        row += 1

    board.refresh()


def _refresh_state(source, engine, delay_sec, game, state):
    """Refresh the game state section of the screen."""
    state.clear()
    state.border()

    state.addstr(1, 2, "CONFIGURATION")
    state.addstr(3, 3, "Players..: %d" % engine.players)
    state.addstr(4, 3, "Mode.....: %s" % engine.mode.value)
    state.addstr(5, 3, "Source...: %s" % type(source).__name__)
    state.addstr(6, 3, "Delay....: %s seconds" % delay_sec)
    state.addstr(7, 3, "State....: %s" % engine.state)

    row = 10
    for player in game.players.values():
        state.addstr(row + 0, 2, "%s PLAYER" % player.color.name)
        state.addstr(row + 2, 3, "Hand.....: %s" % _render_hand(player))
        state.addstr(row + 3, 3, "Pawns....:")
        state.addstr(row + 4, 6, "%s" % player.pawns[0])
        state.addstr(row + 5, 6, "%s" % player.pawns[1])
        state.addstr(row + 6, 6, "%s" % player.pawns[2])
        state.addstr(row + 7, 6, "%s" % player.pawns[3])
        row += 10

    state.refresh()


def _refresh_history(game, history):
    """Refresh the game history section of the screen."""
    history.clear()
    history.border()

    row = 1
    for entry in game.history[-1:]:
        history.addstr(row, 2, "%s" % entry)
        row += 1

    history.refresh()


def _main(stdscr, source: CharacterInputSource, engine: Engine, delay_sec: float, exit_immediately: bool):
    """Main routine for the Python curses application, intended to be wrapped by curses.wrapper()"""

    rows, columns = stdscr.getmaxyx()
    if columns < _MIN_COLS or rows < _MIN_ROWS:
        raise TerminalSizeError("Minimum terminal size is %dx%d, but yours is %dx%d" % (_MIN_COLS, _MIN_ROWS, columns, rows))

    board = curses.newwin(53, 90, 1, 3)
    state = curses.newwin(52, 59, 2, 94)
    history = curses.newwin(3, 150, 54, 3)

    # See https://stackoverflow.com/a/57205676/2907667
    def resize(unused_signum=None, unused_frame=None):
        endwin()
        _draw(stdscr, board, state, history)

    signal(SIGWINCH, resize)
    resize()

    complete = False
    while not complete:  # loop until the user CTRL-C's the application
        if engine.completed:
            if exit_immediately:
                complete = True
                break
        else:
            game = engine.play_next()
            _refresh(source, engine, game, delay_sec, stdscr, board, state, history)
        sleep(delay_sec)


def _force_minimum_size() -> None:
    """Force an xterm to resize via a control sequence."""

    # As of 2020, this worked in both the standard Apple terminal and Debian xterm.
    #
    # In 2022, it no longer works in Debian.  Even in the Apple terminal, it only
    # works if the terminal font and monitor actually allow the requested size, and
    # there's no indication whether it worked or not.
    #
    # I'm apparently using a slightly larger font now than when I originally wrote
    # this code, and these days my terminal can't successfully resize past 155x59 on
    # my Macbook.  The original rendering needed at least 155x70.  To deal with this,
    # I added the TerminalSizeError error handling block (above) to explictly detect
    # that it isn't possible to render the board, and I also adjusted the rendering
    # to work in a slightly smaller terminal.
    #
    # See: https://apple.stackexchange.com/a/47841/249172

    print("\u001b[8;%d;%dt" % (_MIN_ROWS, _MIN_COLS))
    sleep(0.5)  # wait for the window to finish resizing; if we try to render before it's done, the window gets hosed up


def run_demo(players: int, mode: GameMode, source: CharacterInputSource, delay_sec: float, exit_immediately: bool) -> None:
    """
    Run the quick'n'dirty demo in a terminal window.

    Args:
        players(int): Number of players in the game
        mode(GameMode): The game mode
        source(CharacterInputSource): The source to use for choosing player moves
        delay_sec(float): The delay between turns when executing the game
    """
    try:
        characters = [Character(name="Player %d" % player, source=source) for player in range(players)]
        engine = Engine(mode=mode, characters=characters)
        engine.start_game()
        _force_minimum_size()
        curses.wrapper(_main, source, engine, delay_sec, exit_immediately)
    except KeyboardInterrupt:  # users get out using CTRL-C
        print("Demo completed")
        sys.exit(0)
    except TerminalSizeError as e:
        print(e.msg)
        sys.exit(1)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Implementations for command-line (CLI) tools.
"""

import argparse
import sys
from typing import IO, Any, List

from .demo import run_demo
from .game import MAX_PLAYERS, Game, GameMode
from .render import render_board
from .simulation import run_simulation
from .source import source

# Constants used by the demo CLI
_DEMO_DEFAULT_PLAYERS = MAX_PLAYERS
_DEMO_DEFAULT_MODE = GameMode.STANDARD.name
_DEMO_MODE_CHOICES = [GameMode.STANDARD.name, GameMode.ADULT.name]
_DEMO_DEFAULT_SOURCE = "apologies.source.RandomInputSource"
_DEMO_DEFAULT_DELAY_SEC = 1

# Constants used by the simulation CLI
_SIM_DEFAULT_ITERATIONS = 10
_SIM_DEFAULT_OUT = "simulation.csv"


def demo(argv: List[str], unused_stdout: IO[str], unused_stderr: IO[str]) -> None:
    """Run a game with simulated players, displaying output on the terminal."""
    parser = argparse.ArgumentParser(
        description="Run a game with simulated players, displaying output on the terminal.",
        epilog="By default, the game runs in %s mode with %d players. A source is a class that chooses a player's move."
        % (_DEMO_DEFAULT_MODE, _DEMO_DEFAULT_PLAYERS),
    )

    parser.add_argument("--players", type=int, default=_DEMO_DEFAULT_PLAYERS, help="Number of simulated players in the game")
    parser.add_argument("--mode", type=str, default=_DEMO_DEFAULT_MODE, choices=_DEMO_MODE_CHOICES, help="Choose the game mode")
    parser.add_argument("--source", type=str, default=_DEMO_DEFAULT_SOURCE, help="Fully-qualified name of the character source")
    parser.add_argument("--exit", action="store_true", help="Exit immediately when the game completes")
    parser.add_argument(
        "--delay", type=float, default=_DEMO_DEFAULT_DELAY_SEC, help="Delay between computer-generated moves (fractional seconds)"
    )

    args = parser.parse_args(args=argv[2:])
    run_demo(
        players=args.players, mode=GameMode[args.mode], source=source(args.source), delay_sec=args.delay, exit_immediately=args.exit
    )


def simulation(argv: List[str], unused_stdout: IO[str], unused_stderr: IO[str]) -> None:
    """Run a simulation and display results."""
    parser = argparse.ArgumentParser(
        description="Run a simulation to see how well different character input sources behave.",
        epilog="Every combination of game mode, number of players, and input source is tested, "
        "using %d iterations by default.  A spreadsheet is generated containing statistics "
        "on mean and median turns and duration to win, as well as number of wins for each "
        "source." % _SIM_DEFAULT_ITERATIONS,
    )

    parser.add_argument("--iter", type=int, default=_SIM_DEFAULT_ITERATIONS, help="Number of iterations per scenario")

    parser.add_argument("--out", type=str, default=_SIM_DEFAULT_OUT, help="Path to the output CSV file")

    parser.add_argument("source", type=str, nargs="+", help="Fully-qualified name of the character sources to test")

    args = parser.parse_args(args=argv[2:])

    errors = []
    if args.iter <= 0:
        errors.append("simulation: error: there must be at least 1 iteration")

    if errors:
        parser.print_usage()
        print("\n".join(errors))
        sys.exit(1)

    run_simulation(iterations=args.iter, output=args.out, sources=[source(s) for s in args.source])


def render(unused_argv: List[str], stdout: IO[str], unused_stderr: IO[str]) -> None:
    """Render an empty board."""
    game = Game(4)
    board = render_board(game)
    stdout.write("%s" % board)


def example(argv: List[str], stdout: IO[str], stderr: IO[str]) -> None:  # pylint: disable: unused-argument
    """Execute the example script, which just writes some input to its outputs."""
    stdout.write("Hello, stdout: %s\n" % argv[0])
    stderr.write("Hello, stderr: %s\n" % argv[0])


def _lookup_method(method: str) -> Any:
    """Look up the method in this module with the passed-in name."""
    module = sys.modules[__name__]
    return getattr(module, "%s" % method)


def cli(script: str) -> None:
    """
    Run the main routine for the named script.

    Args:
        script(str): Name of the script to execute
    """
    _lookup_method(script)(sys.argv, sys.stdout, sys.stderr)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

from apologies.game import Game, PlayerColor
from apologies.reward import RewardCalculatorV1


class TestRewardCalculatorV1:
    """
    Unit tests for RewardCalculatorV1.
    """

    def test_range(self):
        assert RewardCalculatorV1().range(2) == (0, 400)
        assert RewardCalculatorV1().range(3) == (0, 800)
        assert RewardCalculatorV1().range(4) == (0, 1200)

    def test_empty_game(self):
        for playercount in [2, 3, 4]:
            for color in list(PlayerColor)[:playercount]:
                game = Game(playercount=playercount)
                view = game.create_player_view(color)
                assert RewardCalculatorV1().calculate(view) == 0  # score is always zero if all pawns are in start

    def test_equivalent_state(self):
        game = Game(playercount=4)
        game.players[PlayerColor.RED].pawns[0].position.move_to_square(4)
        game.players[PlayerColor.YELLOW].pawns[0].position.move_to_square(34)
        game.players[PlayerColor.GREEN].pawns[0].position.move_to_square(49)
        game.players[PlayerColor.BLUE].pawns[0].position.move_to_square(19)
        for color in PlayerColor:
            view = game.create_player_view(color)
            assert RewardCalculatorV1().calculate(view) == 0  # score is always zero if all players are equivalent

    def test_safe_zone(self):
        game = Game(playercount=4)
        game.players[PlayerColor.RED].pawns[0].position.move_to_safe(4)  # last safe square before home
        view = game.create_player_view(PlayerColor.RED)
        assert RewardCalculatorV1().calculate(view) == 222
        for color in [PlayerColor.BLUE, PlayerColor.YELLOW, PlayerColor.GREEN]:
            view = game.create_player_view(color)
            assert RewardCalculatorV1().calculate(view) == 0  # score is always zero if all pawns are in start

    def test_winner(self):
        game = Game(playercount=2)
        game.players[PlayerColor.RED].pawns[0].position.move_to_home()
        game.players[PlayerColor.RED].pawns[1].position.move_to_home()
        game.players[PlayerColor.RED].pawns[2].position.move_to_home()
        game.players[PlayerColor.RED].pawns[3].position.move_to_home()
        view = game.create_player_view(PlayerColor.RED)
        assert RewardCalculatorV1().calculate(view) == 400
        for color in [PlayerColor.YELLOW]:
            view = game.create_player_view(color)
            assert RewardCalculatorV1().calculate(view) == 0  # score is always zero if all pawns are in start

        game = Game(playercount=3)
        game.players[PlayerColor.RED].pawns[0].position.move_to_home()
        game.players[PlayerColor.RED].pawns[1].position.move_to_home()
        game.players[PlayerColor.RED].pawns[2].position.move_to_home()
        game.players[PlayerColor.RED].pawns[3].position.move_to_home()
        view = game.create_player_view(PlayerColor.RED)
        assert RewardCalculatorV1().calculate(view) == 800
        for color in [PlayerColor.YELLOW, PlayerColor.GREEN]:
            view = game.create_player_view(color)
            assert RewardCalculatorV1().calculate(view) == 0  # score is always zero if all pawns are in start

        game = Game(playercount=4)
        game.players[PlayerColor.RED].pawns[0].position.move_to_home()
        game.players[PlayerColor.RED].pawns[1].position.move_to_home()
        game.players[PlayerColor.RED].pawns[2].position.move_to_home()
        game.players[PlayerColor.RED].pawns[3].position.move_to_home()
        view = game.create_player_view(PlayerColor.RED)
        assert RewardCalculatorV1().calculate(view) == 1200
        for color in [PlayerColor.YELLOW, PlayerColor.GREEN, PlayerColor.BLUE]:
            view = game.create_player_view(color)
            assert RewardCalculatorV1().calculate(view) == 0  # score is always zero if all pawns are in start

    def test_arbitrary(self):
        game = Game(playercount=4)

        game.players[PlayerColor.RED].pawns[0].position.move_to_home()
        game.players[PlayerColor.RED].pawns[1].position.move_to_safe(0)
        game.players[PlayerColor.RED].pawns[2].position.move_to_square(6)
        game.players[PlayerColor.RED].pawns[3].position.move_to_square(10)

        game.players[PlayerColor.YELLOW].pawns[0].position.move_to_square(34)
        game.players[PlayerColor.YELLOW].pawns[1].position.move_to_square(32)
        game.players[PlayerColor.YELLOW].pawns[2].position.move_to_start()
        game.players[PlayerColor.YELLOW].pawns[3].position.move_to_home()

        game.players[PlayerColor.GREEN].pawns[0].position.move_to_start()
        game.players[PlayerColor.GREEN].pawns[1].position.move_to_start()
        game.players[PlayerColor.GREEN].pawns[2].position.move_to_square(59)
        game.players[PlayerColor.GREEN].pawns[3].position.move_to_start()

        game.players[PlayerColor.BLUE].pawns[0].position.move_to_start()
        game.players[PlayerColor.BLUE].pawns[1].position.move_to_start()
        game.players[PlayerColor.BLUE].pawns[2].position.move_to_start()
        game.players[PlayerColor.BLUE].pawns[3].position.move_to_start()

        view = game.create_player_view(PlayerColor.RED)
        assert RewardCalculatorV1().calculate(view) == 319

        view = game.create_player_view(PlayerColor.YELLOW)
        assert RewardCalculatorV1().calculate(view) == 239

        view = game.create_player_view(PlayerColor.GREEN)
        assert RewardCalculatorV1().calculate(view) == 0

        view = game.create_player_view(PlayerColor.BLUE)
        assert RewardCalculatorV1().calculate(view) == 0



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=redefined-outer-name

import os
from unittest.mock import MagicMock

import pytest

from apologies.cli import _lookup_method, cli, example, render

FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixtures/test_cli")


@pytest.fixture
def data():
    data = {}
    for f in os.listdir(FIXTURE_DIR):
        p = os.path.join(FIXTURE_DIR, f)
        if os.path.isfile(p):
            with open(p, encoding="utf-8") as r:
                data[f] = r.read()
    return data


class TestCli:
    """
    General unit tests for the CLI interface.
    """

    def test_lookup_method(self):
        assert _lookup_method("example") is example
        assert _lookup_method("render") is render
        with pytest.raises(AttributeError):
            assert _lookup_method("")
        with pytest.raises(AttributeError):
            assert _lookup_method("bogus")

    def test_example(self):
        argv = ["1", "a", "b"]
        stdout = MagicMock()
        stderr = MagicMock()
        example(argv, stdout, stderr)
        stdout.write.assert_called_once_with("Hello, stdout: 1\n")
        stderr.write.assert_called_once_with("Hello, stderr: 1\n")

    def test_main(self):
        cli("example")  # just make sure it doesn't blow up


class TestRender:
    """
    Unit tests for the render script.
    """

    def test_render(self, data):
        stdout = MagicMock()
        render(MagicMock(), stdout, MagicMock())
        stdout.write.assert_called_once_with(data["render"])



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=redefined-outer-name,protected-access,broad-except,too-many-public-methods,assigning-non-slot
# Unit tests for engine.py

from unittest.mock import MagicMock, Mock, PropertyMock, call, patch

import pytest

from apologies.engine import Character, Engine
from apologies.game import Card, CardType, GameMode, PlayerColor
from apologies.rules import Action, ActionType, Move, Rules


class TestCharacter:
    def test_constructor(self):
        source = Mock()
        character = Character("c", source)
        assert character.name == "c"
        assert character.source is source

    # noinspection PyTypeChecker
    def test_choose_move_minimal(self):
        source = Mock()
        character = Character("c", source)
        mode = Mock()
        view = Mock()
        legal_moves = []
        evaluator = MagicMock()
        character.choose_move(mode, view, legal_moves, evaluator)
        source.choose_move.assert_called_once_with(mode, view, legal_moves, evaluator)

    # noinspection PyTypeChecker
    def test_choose_move_all_args(self):
        source = Mock()
        character = Character("c", source)
        mode = Mock()
        view = Mock()
        legal_moves = []
        evaluator = MagicMock()
        character.choose_move(mode, view, legal_moves, evaluator)
        source.choose_move.assert_called_once_with(mode, view, legal_moves, evaluator)


class TestEngine:
    def test_constructor_not_random(self):
        character1 = Character("one", Mock())
        character2 = Character("two", Mock())
        engine = Engine(GameMode.STANDARD, [character1, character2], first=PlayerColor.RED)  # 1st player deterministic
        assert engine.players == 2
        assert engine.characters == [character1, character2]
        assert engine.mode == GameMode.STANDARD
        assert engine.started is False
        assert engine.completed is False
        assert engine.state == "Game waiting to start"
        assert engine.game is engine._game
        assert len(engine._game.players) == 2
        assert engine.first == PlayerColor.RED
        assert engine._queue.first == engine.first
        assert engine._queue.entries == [PlayerColor.RED, PlayerColor.YELLOW]
        assert engine._rules.mode == GameMode.STANDARD
        assert engine.colors == engine._map and engine.colors is not engine._map  # it's a copy
        assert engine._map == {PlayerColor.RED: character1, PlayerColor.YELLOW: character2}

    def test_constructor_random(self):
        found = []
        for _ in range(0, 100):
            character1 = Character("one", Mock())
            character2 = Character("two", Mock())
            character3 = Character("three", Mock())
            character4 = Character("four", Mock())
            engine = Engine(GameMode.STANDARD, [character1, character2, character3, character4])  # 1st player random
            assert engine.players == 4
            assert engine.characters == [character1, character2, character3, character4]
            assert engine.mode == GameMode.STANDARD
            assert engine.started is False
            assert engine.completed is False
            assert engine.state == "Game waiting to start"
            assert engine.game is engine._game
            assert len(engine._game.players) == 4
            assert engine.first in [PlayerColor.RED, PlayerColor.YELLOW, PlayerColor.GREEN, PlayerColor.BLUE]
            assert engine._queue.first == engine.first
            assert engine._queue.entries == [PlayerColor.RED, PlayerColor.YELLOW, PlayerColor.GREEN, PlayerColor.BLUE]
            assert engine._rules.mode == GameMode.STANDARD
            assert engine.colors == engine._map and engine.colors is not engine._map  # it's a copy
            assert engine._map == {
                PlayerColor.RED: character1,
                PlayerColor.YELLOW: character2,
                PlayerColor.GREEN: character3,
                PlayerColor.BLUE: character4,
            }
            found.append(engine.first)
        # check that all players showed up as the first player at least once
        assert PlayerColor.RED in found
        assert PlayerColor.YELLOW in found
        assert PlayerColor.BLUE in found
        assert PlayerColor.GREEN in found

    def test_started(self):
        engine = TestEngine._create_engine()
        with patch("apologies.game.Game.started", new_callable=PropertyMock) as started:
            started.return_value = False
            assert engine.started is False
            assert engine.state == "Game waiting to start"
            started.return_value = True
            assert engine.started is True
            assert engine.state == "Game in progress"

    def test_completed(self):
        engine = TestEngine._create_engine()
        with patch("apologies.game.Game.completed", new_callable=PropertyMock) as completed:
            completed.return_value = False
            assert engine.completed is False
            assert engine.state == "Game in progress"
            completed.return_value = True
            assert engine.completed is True
            assert engine.state == "Game completed"

    def test_winner(self):
        engine = TestEngine._create_engine()
        game_winner = MagicMock(color=PlayerColor.YELLOW)
        with patch("apologies.game.Game.completed", new_callable=PropertyMock) as completed:
            with patch("apologies.game.Game.winner", new_callable=PropertyMock) as winner:
                winner.return_value = game_winner
                completed.return_value = False
                assert engine.winner() is None
                completed.return_value = True
                assert engine.winner() == (engine._map[PlayerColor.YELLOW], game_winner)

    def test_reset(self):
        engine = TestEngine._create_engine()
        saved = engine._game
        engine.reset()
        assert engine._game is not None and engine._game is not saved and not engine._game.started

    def test_start_game(self):
        engine = TestEngine._create_engine()
        engine._rules.start_game = MagicMock()
        assert engine.start_game() == engine._game
        engine._rules.start_game.assert_called_once_with(engine._game)

    def test_draw(self):
        engine = TestEngine._create_engine()

        card = MagicMock()
        engine._game.deck.draw = MagicMock(return_value=card)

        assert card is engine.draw()

    def test_discard(self):
        engine = TestEngine._create_engine()

        card = MagicMock()
        engine._game.deck.discard = MagicMock()

        engine.discard(card)
        engine._game.deck.discard.assert_called_once_with(card)

    def test_construct_legal_moves_standard_nocard(self):
        engine = TestEngine._create_engine(mode=GameMode.STANDARD)

        view = Mock()
        drawcard = Mock()
        movecard = Card(0, CardType.CARD_1)
        move = Move(movecard, [])
        legal_moves = [move]
        engine._game.deck.draw = MagicMock(return_value=drawcard)
        engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)

        assert engine.construct_legal_moves(view) == (drawcard, legal_moves)
        engine._rules.construct_legal_moves.assert_called_once_with(view, card=drawcard)

    def test_construct_legal_moves_standard_card(self):
        engine = TestEngine._create_engine(mode=GameMode.STANDARD)

        view = Mock()
        providedcard = Mock()
        drawcard = Mock()
        movecard = Card(0, CardType.CARD_1)
        move = Move(movecard, [])
        legal_moves = [move]
        engine._game.deck.draw = MagicMock(return_value=drawcard)
        engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)

        assert engine.construct_legal_moves(view, card=providedcard) == (providedcard, legal_moves)
        engine._rules.construct_legal_moves.assert_called_once_with(view, card=providedcard)
        engine._game.deck.draw.assert_not_called()

    def test_construct_legal_moves_adult_nocard(self):
        engine = TestEngine._create_engine(mode=GameMode.ADULT)

        view = Mock()
        drawcard = Mock()
        movecard = Card(0, CardType.CARD_1)
        move = Move(movecard, [])
        legal_moves = [move]
        engine._game.deck.draw = MagicMock(return_value=drawcard)
        engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)

        assert engine.construct_legal_moves(view) == (None, legal_moves)
        engine._rules.construct_legal_moves.assert_called_once_with(view, card=None)
        engine._game.deck.draw.assert_not_called()

    def test_construct_legal_moves_adult_card(self):
        engine = TestEngine._create_engine(mode=GameMode.ADULT)

        view = Mock()
        providedcard = Mock()
        drawcard = Mock()
        movecard = Card(0, CardType.CARD_1)
        move = Move(movecard, [])
        legal_moves = [move]
        engine._game.deck.draw = MagicMock(return_value=drawcard)
        engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)

        assert engine.construct_legal_moves(view, card=providedcard) == (providedcard, legal_moves)
        engine._rules.construct_legal_moves.assert_called_once_with(view, card=providedcard)
        engine._game.deck.draw.assert_not_called()

    def test_play_next_completed(self):
        engine = TestEngine._create_engine()
        with patch("apologies.game.Game.completed", new_callable=PropertyMock) as completed:
            completed.return_value = True
            with pytest.raises(ValueError):
                engine.play_next()

    def test_play_next_failed(self):
        engine = TestEngine._create_engine()

        exception = Exception("Hello")
        copy = Mock()

        engine._queue.next = MagicMock(side_effect=exception)
        engine._game.copy = MagicMock(return_value=copy)

        try:
            engine.play_next()
        except Exception:
            assert engine._game is copy  # we replace with a copy of original if method raises exception

    # noinspection PyUnresolvedReferences
    def test_play_next_standard_forfeit(self):
        engine = TestEngine._create_engine()

        card = Card(0, CardType.CARD_1)
        player = engine._game.players[PlayerColor.RED]
        view = Mock()
        move = Move(card, [])
        legal_moves = [move]

        engine._game.create_player_view = MagicMock(return_value=view)
        engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)
        engine._game.track = MagicMock()
        engine._game.deck.draw = MagicMock(return_value=card)
        engine._game.deck.discard = MagicMock()
        engine.characters[0].choose_move = MagicMock(return_value=move)

        engine.play_next()

        engine._game.create_player_view.assert_called_once_with(PlayerColor.RED)
        engine._rules.construct_legal_moves.assert_called_once_with(view, card=card)
        engine.characters[0].choose_move.assert_called_once_with(engine.mode, view, legal_moves, Rules.evaluate_move)
        engine._game.track.assert_called_once_with("Turn is forfeit; discarded card 1", player, card)
        engine._game.deck.discard.assert_called_once_with(card)

    # noinspection PyUnresolvedReferences
    def test_play_next_standard_illegal(self):
        engine = TestEngine._create_engine()

        card = Card(0, CardType.CARD_12)
        player = engine._game.players[PlayerColor.RED]
        pawn = player.pawns[0]
        view = Mock()
        move = Move(card, [Action(ActionType.MOVE_TO_START, Mock())])  # not found in legal_moves
        legal_moves = [Move(card, [Action(ActionType.MOVE_TO_POSITION, pawn)])]

        engine._game.create_player_view = MagicMock(return_value=view)
        engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)
        engine._game.track = MagicMock()
        engine._game.deck.draw = MagicMock(return_value=card)
        engine._game.deck.discard = MagicMock()
        engine.characters[0].choose_move = MagicMock(return_value=move)
        engine._rules.execute_move = MagicMock()
        engine._rules.draw_again = MagicMock(return_value=False)

        engine.play_next()

        engine._game.create_player_view.assert_called_once_with(PlayerColor.RED)
        engine._rules.construct_legal_moves.assert_called_once_with(view, card=card)
        engine.characters[0].choose_move.assert_called_once_with(engine.mode, view, legal_moves, Rules.evaluate_move)
        engine._rules.execute_move.assert_called_once_with(engine._game, player, legal_moves[0])  # we choose random move
        engine._game.deck.discard.assert_called_once_with(card)
        engine._rules.draw_again.assert_called_once_with(card)

    # noinspection PyUnresolvedReferences
    def test_play_next_standard_legal(self):
        engine = TestEngine._create_engine()

        card = Card(0, CardType.CARD_12)
        player = engine._game.players[PlayerColor.RED]
        pawn = player.pawns[0]
        view = Mock()
        move = Move(card, [Action(ActionType.MOVE_TO_START, pawn)])
        legal_moves = [move]

        engine._game.create_player_view = MagicMock(return_value=view)
        engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)
        engine._game.track = MagicMock()
        engine._game.deck.draw = MagicMock(return_value=card)
        engine._game.deck.discard = MagicMock()
        engine.characters[0].choose_move = MagicMock(return_value=move)
        engine._rules.execute_move = MagicMock()
        engine._rules.draw_again = MagicMock(return_value=False)

        engine.play_next()

        engine._game.create_player_view.assert_called_once_with(PlayerColor.RED)
        engine._rules.construct_legal_moves.assert_called_once_with(view, card=card)
        engine.characters[0].choose_move.assert_called_once_with(engine.mode, view, legal_moves, Rules.evaluate_move)
        engine._rules.execute_move.assert_called_once_with(engine._game, player, move)
        engine._game.deck.discard.assert_called_once_with(card)
        engine._rules.draw_again.assert_called_once_with(card)

    # noinspection PyUnresolvedReferences
    def test_play_next_standard_draw_again(self):
        engine = TestEngine._create_engine()

        card1 = Card(0, CardType.CARD_12)
        card2 = Card(1, CardType.CARD_10)
        player = engine._game.players[PlayerColor.RED]
        pawn = player.pawns[0]
        view = Mock()
        move1 = Move(card1, [Action(ActionType.MOVE_TO_START, pawn)])
        move2 = Move(card2, [Action(ActionType.MOVE_TO_POSITION, pawn)])
        legal_moves1 = [move1]
        legal_moves2 = [move2]

        engine._game.create_player_view = MagicMock(return_value=view)
        engine._rules.construct_legal_moves = MagicMock(side_effect=[legal_moves1, legal_moves2])
        engine._game.track = MagicMock()
        engine._game.deck.draw = MagicMock(side_effect=[card1, card2])
        engine._game.deck.discard = MagicMock()
        engine.characters[0].choose_move = MagicMock(side_effect=[move1, move2])
        engine._rules.execute_move = MagicMock()
        engine._rules.draw_again = MagicMock(side_effect=[True, False])

        engine.play_next()

        engine._game.create_player_view.assert_has_calls([call(PlayerColor.RED), call(PlayerColor.RED)])
        engine._rules.construct_legal_moves.assert_has_calls([call(view, card=card1), call(view, card=card2)])
        engine.characters[0].choose_move.assert_has_calls(
            [call(engine.mode, view, legal_moves1, Rules.evaluate_move), call(engine.mode, view, legal_moves2, Rules.evaluate_move)]
        )
        engine._rules.execute_move.assert_has_calls([call(engine._game, player, move1), call(engine._game, player, move2)])
        engine._game.deck.discard.assert_has_calls([call(card1), call(card2)])
        engine._rules.draw_again.assert_has_calls([call(card1), call(card2)])

    # noinspection PyUnresolvedReferences
    def test_play_next_standard_complete(self):
        with patch("apologies.game.Game.completed", new_callable=PropertyMock) as completed:
            completed.side_effect = [False, True]  # not complete when we start execution, but complete after the 1st move

            engine = TestEngine._create_engine()

            card = Card(0, CardType.CARD_12)
            player = engine._game.players[PlayerColor.RED]
            pawn = player.pawns[0]
            view = Mock()
            move = Move(card, [Action(ActionType.MOVE_TO_START, pawn)])
            legal_moves = [move]

            engine._game.create_player_view = MagicMock(return_value=view)
            engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)
            engine._game.track = MagicMock()
            engine._game.deck.draw = MagicMock(return_value=card)
            engine._game.deck.discard = MagicMock()
            engine.characters[0].choose_move = MagicMock(return_value=move)
            engine._rules.execute_move = MagicMock()
            engine._rules.draw_again = MagicMock(return_value=True)  # we won't check this because the complete flag is set

            engine.play_next()

            engine._game.create_player_view.assert_called_once_with(PlayerColor.RED)
            engine._rules.construct_legal_moves.assert_called_once_with(view, card=card)
            engine.characters[0].choose_move.assert_called_once_with(engine.mode, view, legal_moves, Rules.evaluate_move)
            engine._rules.execute_move.assert_called_once_with(engine._game, player, move)
            engine._game.deck.discard.assert_called_once_with(card)
            engine._rules.draw_again.assert_not_called()

    # noinspection PyUnresolvedReferences
    def test_play_next_adult_forfeit(self):
        engine = TestEngine._create_engine(GameMode.ADULT)

        player = engine._game.players[PlayerColor.RED]
        view = Mock()
        movecard = player.hand[0]
        replacementcard = Card(999, CardType.CARD_APOLOGIES)
        move = Move(movecard, [])
        legal_moves = [move]

        engine._game.create_player_view = MagicMock(return_value=view)
        engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)
        engine._game.track = MagicMock()
        engine._game.deck.draw = MagicMock(return_value=replacementcard)
        engine._game.deck.discard = MagicMock()
        engine.characters[0].choose_move = MagicMock(return_value=move)

        engine.play_next()

        engine._game.create_player_view.assert_called_once_with(PlayerColor.RED)
        engine._rules.construct_legal_moves.assert_called_once_with(view, card=None)
        engine.characters[0].choose_move.assert_called_once_with(engine.mode, view, legal_moves, Rules.evaluate_move)
        engine._game.track.assert_called_once_with("Turn is forfeit; discarded card %s" % movecard.cardtype.value, player, movecard)
        engine._game.deck.discard.assert_called_once_with(movecard)

        assert movecard not in player.hand
        assert replacementcard in player.hand

    # noinspection PyUnresolvedReferences
    def test_play_next_adult_illegal(self):
        engine = TestEngine._create_engine(GameMode.ADULT)

        player = engine._game.players[PlayerColor.RED]
        pawn = player.pawns[0]
        view = Mock()
        movecard = player.hand[0]
        replacementcard = Card(999, CardType.CARD_APOLOGIES)
        move = Move(movecard, [Action(ActionType.MOVE_TO_START, Mock())])  # not found in legal_moves
        legal_moves = [Move(movecard, [Action(ActionType.MOVE_TO_POSITION, pawn)])]

        engine._game.create_player_view = MagicMock(return_value=view)
        engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)
        engine._game.track = MagicMock()
        engine._game.deck.draw = MagicMock(return_value=replacementcard)
        engine._game.deck.discard = MagicMock()
        engine.characters[0].choose_move = MagicMock(return_value=move)
        engine._rules.execute_move = MagicMock()
        engine._rules.draw_again = MagicMock(return_value=False)

        engine.play_next()

        engine._game.create_player_view.assert_called_once_with(PlayerColor.RED)
        engine._rules.construct_legal_moves.assert_called_once_with(view, card=None)
        engine.characters[0].choose_move.assert_called_once_with(engine.mode, view, legal_moves, Rules.evaluate_move)
        engine._rules.execute_move.assert_called_once_with(engine._game, player, legal_moves[0])  # we choose random move
        engine._game.deck.discard.assert_called_once_with(movecard)
        engine._rules.draw_again.assert_called_once_with(movecard)

        assert movecard not in player.hand
        assert replacementcard in player.hand

    # noinspection PyUnresolvedReferences
    def test_play_next_adult_legal(self):
        engine = TestEngine._create_engine(GameMode.ADULT)

        player = engine._game.players[PlayerColor.RED]
        pawn = player.pawns[0]
        view = Mock()
        movecard = player.hand[0]
        replacementcard = Card(999, CardType.CARD_APOLOGIES)
        move = Move(movecard, [Action(ActionType.MOVE_TO_START, pawn)])
        legal_moves = [move]

        engine._game.create_player_view = MagicMock(return_value=view)
        engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)
        engine._game.track = MagicMock()
        engine._game.deck.draw = MagicMock(return_value=replacementcard)
        engine._game.deck.discard = MagicMock()
        engine.characters[0].choose_move = MagicMock(return_value=move)
        engine._rules.execute_move = MagicMock()
        engine._rules.draw_again = MagicMock(return_value=False)

        engine.play_next()

        engine._game.create_player_view.assert_called_once_with(PlayerColor.RED)
        engine._rules.construct_legal_moves.assert_called_once_with(view, card=None)
        engine.characters[0].choose_move.assert_called_once_with(engine.mode, view, legal_moves, Rules.evaluate_move)
        engine._rules.execute_move.assert_called_once_with(engine._game, player, move)
        engine._game.deck.discard.assert_called_once_with(movecard)
        engine._rules.draw_again.assert_called_once_with(movecard)

        assert movecard not in player.hand
        assert replacementcard in player.hand

    # noinspection PyUnresolvedReferences
    def test_play_next_adult_draw_again(self):
        engine = TestEngine._create_engine(GameMode.ADULT)

        player = engine._game.players[PlayerColor.RED]
        pawn = player.pawns[0]
        view = Mock()
        movecard1 = player.hand[0]
        movecard2 = player.hand[1]
        replacementcard1 = Card(998, CardType.CARD_APOLOGIES)
        replacementcard2 = Card(999, CardType.CARD_APOLOGIES)
        move1 = Move(movecard1, [Action(ActionType.MOVE_TO_START, pawn)])
        move2 = Move(movecard2, [Action(ActionType.MOVE_TO_POSITION, pawn)])
        legal_moves1 = [move1]
        legal_moves2 = [move2]

        engine._game.create_player_view = MagicMock(return_value=view)
        engine._rules.construct_legal_moves = MagicMock(side_effect=[legal_moves1, legal_moves2])
        engine._game.track = MagicMock()
        engine._game.deck.draw = MagicMock(side_effect=[replacementcard1, replacementcard2])
        engine._game.deck.discard = MagicMock()
        engine.characters[0].choose_move = MagicMock(side_effect=[move1, move2])
        engine._rules.execute_move = MagicMock()
        engine._rules.draw_again = MagicMock(side_effect=[True, False])

        engine.play_next()

        engine._game.create_player_view.assert_has_calls([call(PlayerColor.RED), call(PlayerColor.RED)])
        engine._rules.construct_legal_moves.assert_has_calls([call(view, card=None), call(view, card=None)])
        engine.characters[0].choose_move.assert_has_calls(
            [call(engine.mode, view, legal_moves1, Rules.evaluate_move), call(engine.mode, view, legal_moves2, Rules.evaluate_move)]
        )
        engine._rules.execute_move.assert_has_calls([call(engine._game, player, move1), call(engine._game, player, move2)])
        engine._game.deck.discard.assert_has_calls([call(movecard1), call(movecard2)])
        engine._rules.draw_again.assert_has_calls([call(movecard1), call(movecard2)])

        assert movecard1 not in player.hand
        assert movecard2 not in player.hand
        assert replacementcard1 in player.hand
        assert replacementcard2 in player.hand

    # noinspection PyUnresolvedReferences
    def test_play_next_adult_draw_again_complete(self):
        with patch("apologies.game.Game.completed", new_callable=PropertyMock) as completed:
            completed.side_effect = [False, True]  # not complete when we start execution, but complete after the 1st move

            engine = TestEngine._create_engine(GameMode.ADULT)

            player = engine._game.players[PlayerColor.RED]
            pawn = player.pawns[0]
            view = Mock()
            movecard = player.hand[0]
            replacementcard = Card(999, CardType.CARD_APOLOGIES)
            move = Move(movecard, [Action(ActionType.MOVE_TO_START, pawn)])
            legal_moves = [move]

            engine._game.create_player_view = MagicMock(return_value=view)
            engine._rules.construct_legal_moves = MagicMock(return_value=legal_moves)
            engine._game.track = MagicMock()
            engine._game.deck.draw = MagicMock(return_value=replacementcard)
            engine._game.deck.discard = MagicMock()
            engine.characters[0].choose_move = MagicMock(return_value=move)
            engine._rules.execute_move = MagicMock()
            engine._rules.draw_again = MagicMock(return_value=True)  # we won't check this because the complete flag is set

            engine.play_next()

            engine._game.create_player_view.assert_called_once_with(PlayerColor.RED)
            engine._rules.construct_legal_moves.assert_called_once_with(view, card=None)
            engine.characters[0].choose_move.assert_called_once_with(engine.mode, view, legal_moves, Rules.evaluate_move)
            engine._rules.execute_move.assert_called_once_with(engine._game, player, move)
            engine._game.deck.discard.assert_called_once_with(movecard)
            engine._rules.draw_again.assert_not_called()

            assert movecard not in player.hand
            assert replacementcard in player.hand

    @staticmethod
    def _create_engine(mode: GameMode = GameMode.STANDARD) -> Engine:
        character1 = Character("character1", Mock())
        character1.choose_move = MagicMock()  # type: ignore

        character2 = Character("character2", Mock())
        character1.choose_move = MagicMock()  # type: ignore

        first = PlayerColor.RED
        engine = Engine(mode, [character1, character2], first=first)
        engine.start_game()

        return engine



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Implements rules related to game play.
"""

import uuid
from enum import Enum
from typing import List, Optional

from attrs import define, field, frozen

from .game import (
    ADULT_HAND,
    BOARD_SQUARES,
    CIRCLE,
    DRAW_AGAIN,
    SAFE_SQUARES,
    SLIDE,
    TURN,
    Card,
    CardType,
    Game,
    GameMode,
    Pawn,
    Player,
    PlayerColor,
    PlayerView,
    Position,
)


class ActionType(Enum):
    """Enumeration of all actions that a character can take."""

    MOVE_TO_START = "Move to start"  # Move a pawn back to its start area
    MOVE_TO_POSITION = "Move to position"  # Move a pawn to a specific position on the board


@frozen
class Action:
    # noinspection PyUnresolvedReferences
    """
    An action that can be taken as part of a move.

    Attributes:
        actiontype(ActionType): The type of action
        pawn(Pawn): The pawn that the action operates on
        position(Position): Optionally, a position the pawn should move to
    """

    actiontype: ActionType
    pawn: Pawn
    position: Optional[Position] = None


@frozen
class Move:
    # noinspection PyUnresolvedReferences
    """
    A player's move on the board, which consists of one or more actions.

    Note that the actions associated with a move include both the immediate actions that the player
    chose (such as moving a pawn from start or swapping places with a different pawn), but also
    any side-effects (such as pawns that are bumped back to start because of a slide).  As a result,
    executing a move becomes very easy and no validation is required.  All of the work is done
    up-front.

    Attributes:
        card(Card): The card that is being played by this move
        actions(List[Action]): List of actions to execute
        side_effects(List[Action]): List of side effects that occurred as a result of the actions
        id(str): Identifier for this move, which must be unique among all legal moves this move is grouped with
    """

    card: Card
    actions: List[Action]
    side_effects: List[Action] = field(factory=list)
    id: str = field(factory=lambda: uuid.uuid4().hex)


# noinspection PyMethodMayBeStatic
class BoardRules:

    """
    Rules related to the way the board works.
    """

    # pylint: disable=too-many-return-statements
    def construct_legal_moves(self, color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """
        Return the set of legal moves for a pawn using a card, possibly empty.

        Attributes:
            card(Card): Card to be played
            pawn(Pawn): Pawn that the card will be applied to
            all_pawns(List[Pawn]): All pawns on the board, including the one to be played

        Return:
            Set of legal moves for the pawn using the card.
        """
        moves: List[Move] = []
        if not pawn.position.home:  # there are no legal moves for a pawn in home
            if card.cardtype == CardType.CARD_1:
                moves += BoardRules._construct_legal_moves_1(color, card, pawn, all_pawns)
            elif card.cardtype == CardType.CARD_2:
                moves += BoardRules._construct_legal_moves_2(color, card, pawn, all_pawns)
            elif card.cardtype == CardType.CARD_3:
                moves += BoardRules._construct_legal_moves_3(color, card, pawn, all_pawns)
            elif card.cardtype == CardType.CARD_4:
                moves += BoardRules._construct_legal_moves_4(color, card, pawn, all_pawns)
            elif card.cardtype == CardType.CARD_5:
                moves += BoardRules._construct_legal_moves_5(color, card, pawn, all_pawns)
            elif card.cardtype == CardType.CARD_7:
                moves += BoardRules._construct_legal_moves_7(color, card, pawn, all_pawns)
            elif card.cardtype == CardType.CARD_8:
                moves += BoardRules._construct_legal_moves_8(color, card, pawn, all_pawns)
            elif card.cardtype == CardType.CARD_10:
                moves += BoardRules._construct_legal_moves_10(color, card, pawn, all_pawns)
            elif card.cardtype == CardType.CARD_11:
                moves += BoardRules._construct_legal_moves_11(color, card, pawn, all_pawns)
            elif card.cardtype == CardType.CARD_12:
                moves += BoardRules._construct_legal_moves_12(color, card, pawn, all_pawns)
            elif card.cardtype == CardType.CARD_APOLOGIES:
                moves += BoardRules._construct_legal_moves_apologies(color, card, pawn, all_pawns)
        BoardRules._augment_with_slides(all_pawns, moves)
        return moves

    @staticmethod
    def distance_to_home(pawn: Pawn) -> int:
        """Return the distance to home for this pawn, a number of squares when moving forward."""
        if pawn.position.home:
            return 0
        elif pawn.position.start:
            return 65
        elif pawn.position.safe is not None:
            return SAFE_SQUARES - pawn.position.safe
        else:
            circle = CIRCLE[pawn.color].square
            turn = TURN[pawn.color].square
            square = pawn.position.square
            square_to_corner = BOARD_SQUARES - square  # type: ignore
            corner_to_turn = turn
            turn_to_home = SAFE_SQUARES + 1
            total = square_to_corner + corner_to_turn + turn_to_home  # type: ignore
            if turn < square < circle:  # type: ignore
                return total
            if total < 65:
                return total
            else:
                return total - 60

    # noinspection PyChainedComparisons
    # pylint: disable=too-many-branches,too-many-return-statements,line-too-long
    @staticmethod
    def _position(color: PlayerColor, position: Position, squares: int) -> Position:
        """
        Calculate the new position for a forward or backwards move, taking into account safe zone turns but disregarding slides.
        """
        if position.home or position.start:
            raise ValueError("Pawn in home or start may not move.")
        elif position.safe is not None:
            if squares == 0:
                return position.copy()
            elif squares > 0:
                if position.safe + squares < SAFE_SQUARES:
                    return position.copy().move_to_safe(position.safe + squares)
                elif position.safe + squares == SAFE_SQUARES:
                    return position.copy().move_to_home()
                else:
                    raise ValueError("Pawn cannot move past home.")
            else:  # squares < 0
                if position.safe + squares >= 0:
                    return position.copy().move_to_safe(position.safe + squares)
                else:  # handle moving back out of the safe area
                    return BoardRules._position(color, position.copy().move_to_square(TURN[color].square), squares + position.safe + 1)  # type: ignore
        elif position.square is not None:
            if squares == 0:
                return position.copy()
            elif squares > 0:
                if position.square + squares < BOARD_SQUARES:
                    if position.square <= TURN[color].square and position.square + squares > TURN[color].square:  # type: ignore
                        return BoardRules._position(color, position.copy().move_to_safe(0), squares - (TURN[color].square - position.square) - 1)  # type: ignore
                    else:
                        return position.copy().move_to_square(position.square + squares)
                else:  # handle turning the corner
                    return BoardRules._position(
                        color, position.copy().move_to_square(0), squares - (BOARD_SQUARES - position.square)
                    )
            else:  # squares < 0
                if position.square + squares >= 0:
                    return position.copy().move_to_square(position.square + squares)
                else:  # handle turning the corner
                    return BoardRules._position(
                        color, position.copy().move_to_square(BOARD_SQUARES - 1), squares + position.square + 1
                    )
        else:
            raise ValueError("Position is in an illegal state")

    @staticmethod
    def _construct_legal_moves_1(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_1, possibly empty."""
        moves: List[Move] = []
        moves += BoardRules._move_circle(color, card, pawn, all_pawns)
        moves += BoardRules._move_simple(color, card, pawn, all_pawns, 1)
        return moves

    @staticmethod
    def _construct_legal_moves_2(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_2, possibly empty."""
        moves: List[Move] = []
        moves += BoardRules._move_circle(color, card, pawn, all_pawns)
        moves += BoardRules._move_simple(color, card, pawn, all_pawns, 2)
        return moves

    @staticmethod
    def _construct_legal_moves_3(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_3, possibly empty."""
        return BoardRules._move_simple(color, card, pawn, all_pawns, 3)

    @staticmethod
    def _construct_legal_moves_4(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_4, possibly empty."""
        return BoardRules._move_simple(color, card, pawn, all_pawns, -4)

    @staticmethod
    def _construct_legal_moves_5(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_5, possibly empty."""
        return BoardRules._move_simple(color, card, pawn, all_pawns, 5)

    @staticmethod
    def _construct_legal_moves_7(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_7, possibly empty."""
        moves: List[Move] = []
        moves += BoardRules._move_simple(color, card, pawn, all_pawns, 7)
        moves += BoardRules._move_split(color, card, pawn, all_pawns)
        return moves

    @staticmethod
    def _construct_legal_moves_8(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_8, possibly empty."""
        return BoardRules._move_simple(color, card, pawn, all_pawns, 8)

    @staticmethod
    def _construct_legal_moves_10(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_10, possibly empty."""
        moves: List[Move] = []
        moves += BoardRules._move_simple(color, card, pawn, all_pawns, 10)
        moves += BoardRules._move_simple(color, card, pawn, all_pawns, -1)
        return moves

    @staticmethod
    def _construct_legal_moves_11(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_11, possibly empty."""
        moves: List[Move] = []
        moves += BoardRules._move_swap(color, card, pawn, all_pawns)
        moves += BoardRules._move_simple(color, card, pawn, all_pawns, 11)
        return moves

    @staticmethod
    def _construct_legal_moves_12(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_12, possibly empty."""
        return BoardRules._move_simple(color, card, pawn, all_pawns, 12)

    @staticmethod
    def _construct_legal_moves_apologies(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        """Return the set of legal moves for a pawn using CARD_APOLOGIES, possibly empty."""
        return BoardRules._move_apologies(color, card, pawn, all_pawns)

    @staticmethod
    def _find_pawn(all_pawns: List[Pawn], position: Position) -> Optional[Pawn]:
        """Return the first pawn at the indicated position, or None."""
        for pawn in all_pawns:
            if pawn.position == position:
                return pawn
        return None

    @staticmethod
    def _move_circle(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        # For start-related cards, a pawn in the start area can move to the associated
        # circle position if that position is not occupied by another pawn of the same color.
        moves: List[Move] = []
        if pawn.position.start:
            conflict = BoardRules._find_pawn(all_pawns, CIRCLE[color])
            if not conflict:
                moves.append(Move(card, actions=[Action(ActionType.MOVE_TO_POSITION, pawn, CIRCLE[color].copy())]))
            elif conflict and conflict.color != color:
                moves.append(
                    Move(
                        card,
                        actions=[Action(ActionType.MOVE_TO_POSITION, pawn, CIRCLE[color].copy())],
                        side_effects=[Action(ActionType.MOVE_TO_START, conflict)],
                    )
                )
        return moves

    @staticmethod
    def _move_simple(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn], squares: int) -> List[Move]:
        # For most cards, a pawn on the board can move forward or backward if the
        # resulting position is not occupied by another pawn of the same color.
        moves: List[Move] = []
        if pawn.position.square is not None or pawn.position.safe is not None:
            try:
                target = BoardRules._position(color, pawn.position, squares)
                if target.home or target.start:  # by definition, there can't be a conflict going to home or start
                    moves.append(Move(card, actions=[Action(ActionType.MOVE_TO_POSITION, pawn, target)]))
                else:
                    conflict = BoardRules._find_pawn(all_pawns, target)
                    if not conflict:
                        moves.append(Move(card, actions=[Action(ActionType.MOVE_TO_POSITION, pawn, target)]))
                    elif conflict and conflict.color != color:
                        moves.append(
                            Move(
                                card,
                                actions=[Action(ActionType.MOVE_TO_POSITION, pawn, target)],
                                side_effects=[Action(ActionType.MOVE_TO_START, conflict)],
                            )
                        )
            except ValueError as ignored:
                pass  # if the requested position is not legal, then just ignore it
        return moves

    @staticmethod
    def _move_split(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        # For the 7 card, we can split up the move between two different pawns.
        # Any combination of 7 forward moves is legal, as long as the resulting position
        # is not occupied by another pawn of the same color.
        moves: List[Move] = []
        for other in all_pawns:
            if other != pawn and other.color == color and not other.position.home and not other.position.start:
                for (left, right) in [(1, 6), (2, 5), (3, 4), (4, 3), (5, 2), (6, 1)]:  # legal ways to split up a move of 7
                    left_moves = BoardRules._move_simple(color, card, pawn, [p for p in all_pawns if p != other], left)
                    right_moves = BoardRules._move_simple(color, card, other, [p for p in all_pawns if p != other], right)
                    if left_moves and right_moves:
                        moves.append(
                            Move(
                                card,
                                actions=left_moves[0].actions + right_moves[0].actions,
                                side_effects=left_moves[0].side_effects + right_moves[0].side_effects,
                            )
                        )
        return moves

    @staticmethod
    def _move_swap(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        # For the 11 card, a pawn on the board can swap with another pawn of a different
        # color, as long as that pawn is outside of the start area, safe area, or home area.
        moves: List[Move] = []
        if pawn.position.square is not None:  # pawn is on the board
            for swap in all_pawns:
                if swap.color != color and not swap.position.home and not swap.position.start and swap.position.safe is None:
                    moves.append(
                        Move(
                            card,
                            actions=[
                                Action(ActionType.MOVE_TO_POSITION, pawn, swap.position.copy()),
                                Action(ActionType.MOVE_TO_POSITION, swap, pawn.position.copy()),
                            ],
                        )
                    )
        return moves

    @staticmethod
    def _move_apologies(color: PlayerColor, card: Card, pawn: Pawn, all_pawns: List[Pawn]) -> List[Move]:
        # For the Apologies card, a pawn in start can swap with another pawn of a different
        # color, as long as that pawn is outside of the start area, safe area, or home area.
        moves: List[Move] = []
        if pawn.position.start:
            for swap in all_pawns:
                if swap.color != color and not swap.position.home and not swap.position.start and swap.position.safe is None:
                    moves.append(
                        Move(
                            card,
                            actions=[
                                Action(ActionType.MOVE_TO_POSITION, pawn, swap.position.copy()),
                                Action(ActionType.MOVE_TO_START, swap),
                            ],
                        )
                    )
        return moves

    # pylint: disable=too-many-nested-blocks
    @staticmethod
    def _augment_with_slides(all_pawns: List[Pawn], moves: List[Move]) -> None:
        """Augument any legal moves with additional side-effects that occur as a result of slides."""
        for move in moves:
            for action in move.actions:
                if action.actiontype == ActionType.MOVE_TO_POSITION:  # look at any move to a position on the board
                    for color in [color for color in PlayerColor if color != action.pawn.color]:  # any color other than the pawn's
                        for (start, end) in SLIDE[color]:  # look at all slides with this color
                            if action.position and action.position.square == start:  # if the pawn landed on the start of the slide
                                action.position.move_to_square(end)  # move the pawn to the end of the slide
                                for square in range(start + 1, end + 1):  # and then bump any pawns that were already on the slide
                                    # Note: in this one case, a pawn can bump another pawn of the same color
                                    pawn = BoardRules._find_pawn(all_pawns, Position().move_to_square(square))
                                    if pawn:
                                        bump = Action(ActionType.MOVE_TO_START, pawn)
                                        if bump not in move.actions:
                                            move.side_effects.append(bump)


# noinspection PyProtectedMember
@define(slots=False)
class Rules:
    # noinspection PyUnresolvedReferences
    """
    Implements rules related to game play.

    Attributes:
        mode(GameMode): The game mode
    """

    mode: GameMode
    _board_rules: BoardRules = field(init=False, factory=BoardRules)

    # noinspection PyMethodMayBeStatic
    def draw_again(self, card: Card) -> bool:
        """Whether the player gets to draw again based on the passed-in card."""
        return DRAW_AGAIN[card.cardtype]

    def start_game(self, game: Game) -> None:
        """
        Start the game.

        Args:
            game(Game): Game to operate on
        """
        if game.started:
            raise ValueError("Game is already started")
        game.track("Game started with mode: %s" % self.mode)
        if self.mode == GameMode.ADULT:
            Rules._setup_adult_mode(game)

    def construct_legal_moves(self, view: PlayerView, card: Optional[Card] = None) -> List[Move]:
        """
        Return the set of all legal moves for a player and its opponents.

        Attributes:
            view(PlayerView): Player-specific view of the game
            card(Card, optional): The card to play, or None if move should come from player's hand

        Returns:
            List[Move]: Set of legal moves for the player, as described above.
        """
        moves: List[Move] = []
        all_pawns = view.all_pawns()
        for played in [card] if card else view.player.hand:
            for pawn in view.player.pawns:
                for move in self._board_rules.construct_legal_moves(view.player.color, played, pawn, all_pawns):
                    if move not in moves:  # filter out duplicates
                        moves.append(move)
        if not moves:  # if there are no legal moves, then forfeit (discarding one card) becomes the only allowable move
            for played in [card] if card else view.player.hand:
                moves.append(Move(played, []))
        if not moves:  # if there are still no legal moves, then this is an internal error
            raise ValueError("Internal error: could not construct any legal moves")
        return moves

    # noinspection PyMethodMayBeStatic
    def execute_move(self, game: Game, player: Player, move: Move) -> None:
        """
        Execute a player's move, updating game state.

        Args:
            game(Game): Game to operate on
            player(Player): Color of the player associated with the move
            move(Move): Move to validate
        """
        log = "Played card %s: [ " % move.card.cardtype.value
        for action in move.actions + move.side_effects:  # execute actions, then side-effects, in order
            # Keep in mind that the pawn on the action is a different object than the pawn in the game
            pawn = game.players[action.pawn.color].pawns[action.pawn.index]
            if action.actiontype == ActionType.MOVE_TO_START:
                pawn.position.move_to_start()
                log += "%s->start, " % pawn.name
            elif action.actiontype == ActionType.MOVE_TO_POSITION and action.position:
                pawn.position.move_to_position(action.position)
                log += "%s, " % pawn
        log += "]"
        game.track(log, player, move.card)
        if game.completed:
            game.track("Game completed: winner is %s after %d turns" % (game.winner.color.value, game.winner.turns))  # type: ignore

    @staticmethod
    def evaluate_move(view: PlayerView, move: Move) -> PlayerView:
        """
        Construct the new player view that results from executing the passed-in move.

        This is equivalent to execute_move() but has no permanent effect on the game.  It's intended for
        use by a character, to evaluate the results of each legal move.

        Args:
            view(PlayerView):
            move(Move):

        Returns:
            PlayerView: The new state after executing the move.
        """
        result = view.copy()
        for action in move.actions + move.side_effects:  # execute actions, then side-effects, in order
            # Keep in mind that the pawn on the action is a different object than the pawn in the view
            pawn = result.get_pawn(action.pawn)
            if pawn:  # if the pawn isn't valid, just ignore it
                if action.actiontype == ActionType.MOVE_TO_START:
                    pawn.position.move_to_start()
                elif action.actiontype == ActionType.MOVE_TO_POSITION and action.position:
                    pawn.position.move_to_position(action.position)
        return result

    @staticmethod
    def _setup_adult_mode(game: Game) -> None:
        """Setup adult mode at the start of the game, which moves some pieces and deals some cards."""
        for player in game.players.values():
            player.pawns[0].position.move_to_position(CIRCLE[player.color])
        for _ in range(ADULT_HAND):
            for player in game.players.values():
                player.hand.append(game.deck.draw())



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Utility functionality.
"""
from typing import Generic, List, TypeVar

import cattrs
from attrs import define, field
from pendulum.datetime import DateTime
from pendulum.parser import parse


class CattrConverter(cattrs.GenConverter):
    """
    Cattr converter that knows how to correctly serialize/deserialize DateTime to an ISO 8601 timestamp.
    """

    # Note: we need to use GenConverter and not Converter because we use PEP563 (postponed) annotations
    # See: https://stackoverflow.com/a/72539298/2907667 and https://github.com/python-attrs/cattrs/issues/41

    def __init__(self) -> None:
        super().__init__()
        self.register_unstructure_hook(DateTime, lambda datetime: datetime.isoformat() if datetime else None)  # type: ignore
        self.register_structure_hook(DateTime, lambda string, _: parse(string) if string else None)


T = TypeVar("T")  # pylint: disable=invalid-name
"""Generic type"""


@define(slots=False)
class CircularQueue(Generic[T]):
    """A circular queue that keeps returning the original entries repeatedly, in order."""

    entries: List[T] = field()
    first: T = field()
    _working: List[T] = field()

    @first.default
    def _default_first(self) -> T:
        if not self.entries:
            raise ValueError("Entries must not be empty")
        return self.entries[0]

    # noinspection PyUnresolvedReferences
    @_working.default
    def _default_working(self) -> List[T]:
        if self.first not in self.entries:
            raise ValueError("First entry not found")
        temp = self.entries[:]
        popped = temp.pop(0)
        while popped != self.first:
            popped = temp.pop(0)
        temp.insert(0, popped)
        return temp

    def next(self) -> T:
        """Get the next entry from the queue."""
        if len(self._working) == 0:
            self._working.extend(self.entries)
        return self._working.pop(0)
